"""Internal utilities"""

import re
from typing import Union


class AttrDict(dict):
    """Dictionary that allows accessing values using the "dot notation" as attributes."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            {_camel2snake(key): to_attrdict(value) for key, value in dict(*args, **kwargs).items()}
        )

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]
        raise AttributeError("Attribute not found: " + str(attr))

    def __setattr__(self, attr, val):
        if attr in self:
            raise AttributeError("Attribute-style access is read only")
        super().__setattr__(attr, val)


def _camel2snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("__([A-Z])", r"_\1", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()


def _snake2camel(obj: dict) -> dict:
    return {_snake2camel_str(key): value for key, value in obj.items()}


def _snake2camel_str(name: str) -> str:
    parts = name.split("_")
    return parts[0] + "".join(word.capitalize() for word in parts[1:])


def to_attrdict(obj: Union[AttrDict, dict, list]) -> Union[AttrDict, list]:
    """Convert any dict or dict inside a list to AttrDict"""
    if isinstance(obj, AttrDict):
        return obj
    if isinstance(obj, dict):
        return AttrDict(obj)
    if isinstance(obj, list):
        return [to_attrdict(elem) for elem in obj]
    return obj



"""Minimal/simple single-account client example boilerplate."""

import subprocess
import sys

from deltachat2 import Client, CoreEvent, IOTransport, Rpc, events

hooks = events.HookCollection()


@hooks.on(events.RawEvent)
def log_event(_client: Client, accid: int, event: CoreEvent) -> None:
    """here you should process events and update UI."""
    print(f"[accid={accid}] {event}")


def main() -> None:
    """Configure (if necessary) and run the client."""
    # in a TUI app you don't want the deltachat-rpc-server printing to stderr
    with IOTransport(stderr=subprocess.DEVNULL) as trans:
        rpc = Rpc(trans)
        client = Client(rpc, hooks)

        accounts = rpc.get_all_account_ids()
        accid = accounts[0] if accounts else rpc.add_account()

        if not rpc.is_configured(accid):
            if len(sys.argv) != 3:
                print("ERROR: Account is not configured so email and password must be provided")
                return
            client.configure(accid, sys.argv[1], sys.argv[2])

        client.run_forever()


if __name__ == "__main__":
    main()



"""Event loop implementations offering high level event handling/hooking for bots."""

import logging
from typing import Callable, Iterable, Optional, Tuple, Union

from .client import Client
from .events import EventFilter, HookCallback, NewMessage
from .rpc import Rpc
from .transport import JsonRpcError
from .types import Event, EventType, Message, NewMsgEvent, SpecialContactId


class Bot(Client):
    """A Delta Chat client with the bot setting set to 1.

    This bot client triggers "NewMessage" highlevel events
    in addition to raw core events. The account will have the settings bcc_self set to 0,
    and delete_server_after set to 1.
    """

    def __init__(
        self,
        rpc: Rpc,
        hooks: Optional[Iterable[Tuple[HookCallback, Union[type, EventFilter]]]] = None,
        logger: Optional[logging.Logger] = None,
        command_prefix: str = "/",
    ) -> None:
        """If hooks is an instance of HookCollection, also its post-hooks will be registered."""
        self.command_prefix = command_prefix
        logger = logger or logging.getLogger("deltachat2.Bot")
        super().__init__(rpc, hooks, logger)

    def configure(self, account_id: int, email: str, password: str, **kwargs) -> None:
        """Configure the account with the given account ID."""
        kwargs.setdefault("bot", "1")
        kwargs.setdefault("delete_server_after", "1")
        kwargs.setdefault("bcc_self", "0")
        super().configure(account_id, email, password, **kwargs)

    def has_command(self, command: str) -> bool:
        """Return True if the bot has a hook/callback registered for the given command,
        False otherwise."""
        if not command or not command.startswith(self.command_prefix):
            return False
        for hook in self._hooks.get(NewMessage, []):
            if command == hook[1].command:
                return True
        return False

    def run_until(self, func: Callable[[Event], bool], account_id: int = 0) -> Event:
        """Process events until the given callable evaluates to True.

        The callable will receive the Event object representing the last processed event.
        The event is returned when the callable evaluates to True.
        """
        if account_id:
            if self.rpc.is_configured(account_id):
                self.rpc.start_io(account_id)
                self._process_messages(account_id)  # Process old messages.
        else:
            self.rpc.start_io_for_all_accounts()
            for acc_id in self.rpc.get_all_account_ids():
                if self.rpc.is_configured(acc_id):
                    self._process_messages(acc_id)  # Process old messages.

        def _wrapper(event: Event) -> bool:
            kind = event.event.kind
            if kind in (EventType.INCOMING_MSG, EventType.MSGS_CHANGED):
                self._process_messages(event.account_id)
            return func(event)

        return super().run_until(_wrapper, account_id)

    def _parse_command(self, accid: int, event: NewMsgEvent) -> None:
        cmds = [hook[1].command for hook in self._hooks.get(NewMessage, []) if hook[1].command]
        parts = event.msg.text.split(maxsplit=1)
        payload = parts[1] if len(parts) > 1 else ""
        cmd = parts.pop(0)

        if "@" in cmd:
            suffix = "@" + self.rpc.get_contact(accid, SpecialContactId.SELF).address
            if cmd.endswith(suffix):
                cmd = cmd[: -len(suffix)]
            else:
                return

        parts = cmd.split("_")
        _payload = payload
        while parts:
            _cmd = "_".join(parts)
            if _cmd in cmds:
                break
            _payload = (parts.pop() + " " + _payload).rstrip()

        if parts:
            cmd = _cmd
            payload = _payload

        event.command = cmd
        event.payload = payload

    def _on_new_msg(self, accid: int, msg: Message) -> None:
        event = NewMsgEvent(command="", payload="", msg=msg)
        if not msg.is_info and msg.text.startswith(self.command_prefix):
            self._parse_command(accid, event)
        self._on_event(Event(accid, event), NewMessage)  # noqa

    def _process_messages(self, accid: int, retry=True) -> None:
        try:
            for msgid in self.rpc.get_next_msgs(accid):
                msg = self.rpc.get_message(accid, msgid)
                outgoing = msg.from_id == SpecialContactId.SELF
                if outgoing or msg.from_id > SpecialContactId.LAST_SPECIAL:
                    self._on_new_msg(accid, msg)
                self.rpc.set_config(accid, "last_msg_id", str(msgid))
        except JsonRpcError as err:
            self.logger.exception(err)
            if retry:
                self._process_messages(accid, False)



"""Advanced single-account echo-bot example."""

import sys

from deltachat2 import (
    Bot,
    CoreEvent,
    EventType,
    IOTransport,
    MsgData,
    NewMsgEvent,
    Rpc,
    events,
)

hooks = events.HookCollection()


@hooks.on(events.RawEvent)
def log_event(bot: Bot, accid: int, event: CoreEvent) -> None:
    """Log all core events for debugging."""
    if event.kind == EventType.INFO:
        bot.logger.debug(event.msg)
    elif event.kind == EventType.WARNING:
        bot.logger.warning(event.msg)
    elif event.kind == EventType.ERROR:
        bot.logger.error(event.msg)
    elif event.kind == EventType.MSG_DELIVERED:
        bot.rpc.delete_messages(accid, [event.msg_id])


@hooks.on(events.NewMessage)
def echo(bot: Bot, accid: int, event: NewMsgEvent) -> None:
    """Echo back any text message"""
    msg = event.msg
    bot.rpc.markseen_msgs(accid, [msg.id])
    bot.rpc.send_msg(accid, msg.chat_id, MsgData(text=msg.text))


@hooks.after(events.NewMessage)
def delete_msgs(bot: Bot, accid: int, event: NewMsgEvent) -> None:
    """Delete already processed messages."""
    bot.rpc.delete_messages(accid, [event.msg.id])


def main() -> None:
    """Configure (if necessary) and run the bot."""
    with IOTransport() as trans:
        rpc = Rpc(trans)
        bot = Bot(rpc, hooks)

        accounts = rpc.get_all_account_ids()
        accid = accounts[0] if accounts else rpc.add_account()

        print("Running deltachat core", rpc.get_system_info().deltachat_core_version)

        if not rpc.is_configured(accid):
            if len(sys.argv) != 3:
                print("ERROR: Account is not configured so email and password must be provided")
                return
            bot.configure(accid, sys.argv[1], sys.argv[2])

        bot.run_forever()


if __name__ == "__main__":
    main()



"""Data classes and types from the JSON-RPC."""

from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Optional, Tuple

from ._utils import AttrDict


class ContactFlag(IntEnum):
    """Contact flags used for filtering"""

    VERIFIED_ONLY = 0x01
    ADD_SELF = 0x02


class ChatlistFlag(IntEnum):
    """Chatlist flags used for filtering"""

    ARCHIVED_ONLY = 0x01
    NO_SPECIALS = 0x02
    ADD_ALLDONE_HINT = 0x04
    FOR_FORWARDING = 0x08


class SpecialContactId(IntEnum):
    """Special contact IDs"""

    SELF = 1
    INFO = 2  # centered messages as "member added", used in all chats
    DEVICE = 5  #  messages "update info" in the device-chat
    LAST_SPECIAL = 9


class EventType(str, Enum):
    """Core event types"""

    INFO = "Info"
    SMTP_CONNECTED = "SmtpConnected"
    IMAP_CONNECTED = "ImapConnected"
    SMTP_MESSAGE_SENT = "SmtpMessageSent"
    IMAP_MESSAGE_DELETED = "ImapMessageDeleted"
    IMAP_MESSAGE_MOVED = "ImapMessageMoved"
    IMAP_INBOX_IDLE = "ImapInboxIdle"
    NEW_BLOB_FILE = "NewBlobFile"
    DELETED_BLOB_FILE = "DeletedBlobFile"
    WARNING = "Warning"
    ERROR = "Error"
    ERROR_SELF_NOT_IN_GROUP = "ErrorSelfNotInGroup"
    MSGS_CHANGED = "MsgsChanged"
    REACTIONS_CHANGED = "ReactionsChanged"
    INCOMING_MSG = "IncomingMsg"
    INCOMING_MSG_BUNCH = "IncomingMsgBunch"
    MSGS_NOTICED = "MsgsNoticed"
    MSG_DELIVERED = "MsgDelivered"
    MSG_FAILED = "MsgFailed"
    MSG_READ = "MsgRead"
    MSG_DELETED = "MsgDeleted"
    CHAT_MODIFIED = "ChatModified"
    CHAT_EPHEMERAL_TIMER_MODIFIED = "ChatEphemeralTimerModified"
    CONTACTS_CHANGED = "ContactsChanged"
    LOCATION_CHANGED = "LocationChanged"
    CONFIGURE_PROGRESS = "ConfigureProgress"
    IMEX_PROGRESS = "ImexProgress"
    IMEX_FILE_WRITTEN = "ImexFileWritten"
    SECUREJOIN_INVITER_PROGRESS = "SecurejoinInviterProgress"
    SECUREJOIN_JOINER_PROGRESS = "SecurejoinJoinerProgress"
    CONNECTIVITY_CHANGED = "ConnectivityChanged"
    SELFAVATAR_CHANGED = "SelfavatarChanged"
    WEBXDC_STATUS_UPDATE = "WebxdcStatusUpdate"
    WEBXDC_INSTANCE_DELETED = "WebxdcInstanceDeleted"


class ChatType(IntEnum):
    """Chat types"""

    UNDEFINED = 0
    SINGLE = 100
    GROUP = 120
    MAILINGLIST = 140
    BROADCAST = 160


class ChatVisibility(str, Enum):
    """Chat visibility types"""

    NORMAL = "Normal"
    ARCHIVED = "Archived"
    PINNED = "Pinned"


class DownloadState(str, Enum):
    """Message download state"""

    DONE = "Done"
    AVAILABLE = "Available"
    FAILURE = "Failure"
    IN_PROGRESS = "InProgress"


class MessageViewtype(str, Enum):
    """Message view type."""

    UNKNOWN = "Unknown"
    TEXT = "Text"
    IMAGE = "Image"
    GIF = "Gif"
    STICKER = "Sticker"
    AUDIO = "Audio"
    VOICE = "Voice"
    VIDEO = "Video"
    FILE = "File"
    VIDEOCHAT_INVITATION = "VideochatInvitation"
    WEBXDC = "Webxdc"


class SystemMessageType(str, Enum):
    """System message type."""

    UNKNOWN = "Unknown"
    GROUP_NAME_CHANGED = "GroupNameChanged"
    GROUP_IMAGE_CHANGED = "GroupImageChanged"
    MEMBER_ADDED_TO_GROUP = "MemberAddedToGroup"
    MEMBER_REMOVED_FROM_GROUP = "MemberRemovedFromGroup"
    AUTOCRYPT_SETUP_MESSAGE = "AutocryptSetupMessage"
    SECUREJOIN_MESSAGE = "SecurejoinMessage"
    LOCATION_STREAMING_ENABLED = "LocationStreamingEnabled"
    LOCATION_ONLY = "LocationOnly"
    CHAT_PROTECTION_ENABLED = "ChatProtectionEnabled"
    CHAT_PROTECTION_DISABLED = "ChatProtectionDisabled"
    WEBXDC_STATUS_UPDATE = "WebxdcStatusUpdate"
    EPHEMERAL_TIMER_CHANGED = "EphemeralTimerChanged"
    MULTI_DEVICE_SYNC = "MultiDeviceSync"
    WEBXDC_INFO_MESSAGE = "WebxdcInfoMessage"


class MessageState(IntEnum):
    """State of the message."""

    UNDEFINED = 0
    IN_FRESH = 10
    IN_NOTICED = 13
    IN_SEEN = 16
    OUT_PREPARING = 18
    OUT_DRAFT = 19
    OUT_PENDING = 20
    OUT_FAILED = 24
    OUT_DELIVERED = 26
    OUT_MDN_RCVD = 28


class SpecialMessageId(IntEnum):
    """Special message ids"""

    DAYMARKER = 9
    LAST_SPECIAL = 9


class CertificateChecks(IntEnum):
    """Certificate checks mode"""

    AUTOMATIC = 0
    STRICT = 1
    ACCEPT_INVALID_CERTIFICATES = 3


class Connectivity(IntEnum):
    """Connectivity states"""

    NOT_CONNECTED = 1000
    CONNECTING = 2000
    WORKING = 3000
    CONNECTED = 4000


class KeyGenType(IntEnum):
    """Type of the key to generate"""

    DEFAULT = 0
    RSA2048 = 1
    ED25519 = 2
    RSA4096 = 3


# "Lp" means "login parameters"
class LpAuthFlag(IntEnum):
    """Authorization flags"""

    OAUTH2 = 0x2
    NORMAL = 0x4


class MediaQuality(IntEnum):
    """Media quality setting"""

    BALANCED = 0
    WORSE = 1


class ProviderStatus(IntEnum):
    """Provider status according to manual testing"""

    OK = 1
    PREPARATION = 2
    BROKEN = 3


class PushNotifyState(IntEnum):
    """Push notifications state"""

    NOT_CONNECTED = 0
    HEARTBEAT = 1
    CONNECTED = 2


class ShowEmails(IntEnum):
    """Show emails mode"""

    OFF = 0
    ACCEPTED_CONTACTS = 1
    ALL = 2


class SocketSecurity(IntEnum):
    """Socket security"""

    AUTOMATIC = 0
    SSL = 1
    STARTTLS = 2
    PLAIN = 3


class VideochatType(IntEnum):
    """Video chat URL type"""

    UNKNOWN = 0
    BASICWEBRTC = 1
    JITSI = 2


class CoreEvent(AttrDict):
    """Delta Chat core Event"""

    kind: EventType


@dataclass
class Event:
    """Low level RPC event."""

    account_id: int
    event: CoreEvent


@dataclass
class NewMsgEvent:
    """New message bot-specific event"""

    command: str
    payload: str
    msg: "Message"


@dataclass
class MsgData:
    """Message data provided to Rpc.send_msg()"""

    text: Optional[str] = None
    html: Optional[str] = None
    viewtype: Optional[MessageViewtype] = None
    file: Optional[str] = None
    location: Optional[Tuple[float, float]] = None
    override_sender_name: Optional[str] = None
    quoted_message_id: Optional[int] = None


class Message(AttrDict):
    """Message snapshot"""



"""JSON-RPC transports to communicate with Delta Chat core."""

import itertools
import json
import logging
import os
import subprocess
import sys
from abc import ABC, abstractmethod
from queue import Queue
from threading import Event, Thread
from typing import Any, Dict, Iterator, Optional

from ._utils import to_attrdict


class JsonRpcError(Exception):
    """An error occurred in your request to the JSON-RPC API."""


class RpcTransport(ABC):
    """Delta Chat RPC client's transport."""

    @abstractmethod
    def call(self, method: str, *args) -> Any:
        """Request the RPC server to call a function and return its return value if any."""


class _Result(Event):
    def __init__(self) -> None:
        self._value: Any = None
        super().__init__()

    def set(self, value: Any) -> None:  # noqa
        self._value = value
        super().set()

    def wait(self) -> Any:  # noqa
        super().wait()
        return self._value


class IOTransport:
    """Delta Chat RPC transport over IO using external deltachat-rpc-server program."""

    def __init__(self, accounts_dir: Optional[str] = None, **kwargs):
        """The given arguments will be passed to subprocess.Popen()"""
        self.logger = logging.getLogger("deltachat2.IOTransport")
        if accounts_dir:
            kwargs["env"] = {
                **kwargs.get("env", os.environ),
                "DC_ACCOUNTS_PATH": str(accounts_dir),
            }

        self._kwargs = kwargs
        self.process: subprocess.Popen
        self.id_iterator: Iterator[int]
        # Map from request ID to the result.
        self.pending_results: Dict[int, _Result]
        self.request_queue: Queue
        self.closing: bool
        self.reader_thread: Thread
        self.writer_thread: Thread

    def start(self) -> None:
        """Start the RPC server process."""
        if sys.version_info >= (3, 11):
            # Prevent subprocess from capturing SIGINT.
            kwargs = {"process_group": 0, **self._kwargs}
        else:
            # `process_group` is not supported before Python 3.11.
            kwargs = {"preexec_fn": os.setpgrp, **self._kwargs}  # noqa: PLW1509
        self.process = subprocess.Popen(  # noqa: R1732
            "deltachat-rpc-server",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            **kwargs,
        )
        self.id_iterator = itertools.count(start=1)
        self.pending_results = {}
        self.request_queue = Queue()
        self.closing = False
        self.reader_thread = Thread(target=self._reader_loop)
        self.reader_thread.start()
        self.writer_thread = Thread(target=self._writer_loop)
        self.writer_thread.start()

    def close(self) -> None:
        """Terminate RPC server process and wait until the reader loop finishes."""
        self.closing = True
        self.call("stop_io_for_all_accounts")
        assert self.process.stdin
        self.process.stdin.close()
        self.reader_thread.join()
        self.request_queue.put(None)
        self.writer_thread.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, _exc_type, _exc, _tb):
        self.close()

    def _reader_loop(self) -> None:
        try:
            assert self.process.stdout
            while True:
                line = self.process.stdout.readline()
                if not line:  # EOF
                    break
                response = json.loads(line)
                if "id" in response:
                    self.pending_results.pop(response["id"]).set(response)
                else:
                    self.logger.warning("Got a response without ID: %s", response)
        except Exception:
            # Log an exception if the reader loop dies.
            self.logger.exception("Exception in the reader loop")

    def _writer_loop(self) -> None:
        """Writer loop ensuring only a single thread writes requests."""
        try:
            assert self.process.stdin
            while True:
                request = self.request_queue.get()
                if not request:
                    break
                data = (json.dumps(request) + "\n").encode()
                self.process.stdin.write(data)
                self.process.stdin.flush()
        except Exception:
            # Log an exception if the writer loop dies.
            self.logger.exception("Exception in the writer loop")

    def call(self, method: str, *args) -> Any:
        """Request the RPC server to call a function and return its return value if any."""
        request_id = next(self.id_iterator)
        request = {
            "jsonrpc": "2.0",
            "method": method,
            "params": args,
            "id": request_id,
        }
        result = self.pending_results[request_id] = _Result()
        self.request_queue.put(request)
        response = result.wait()

        if "error" in response:
            raise JsonRpcError(response["error"])
        if "result" in response:
            return to_attrdict(response["result"])
        return None



"""High-level classes for event processing and filtering."""

import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Set,
    Tuple,
    Union,
)

from .types import CoreEvent, EventType, NewMsgEvent, SpecialContactId

if TYPE_CHECKING:
    from .bot import Bot
    from .client import Client
ClientHookCallback = Callable[["Client", int, CoreEvent], None]
BotHookCallback = Union[
    Callable[["Bot", int, CoreEvent], None], Callable[["Bot", int, NewMsgEvent], None]
]
HookCallback = Union[ClientHookCallback, BotHookCallback]
HookDecorator = Callable[[HookCallback], HookCallback]


def _tuple_of(obj, type_: type) -> tuple:
    if not obj:
        return ()
    if isinstance(obj, type_):
        obj = (obj,)

    if not all(isinstance(elem, type_) for elem in obj):
        raise TypeError()
    return tuple(obj)


class EventFilter(ABC):
    """The base event filter."""

    def __init__(self, func: Optional[Callable] = None):
        self.func = func

    @abstractmethod
    def __hash__(self) -> int:
        """Object's unique hash"""

    @abstractmethod
    def __eq__(self, other) -> bool:
        """Return True if two event filters are equal."""

    def __ne__(self, other):
        return not self == other

    def _call_func(self, event: Union[CoreEvent, NewMsgEvent]) -> bool:
        if not self.func:
            return True
        return bool(self.func(event))

    @abstractmethod
    def filter(self, event: Union[CoreEvent, NewMsgEvent]):
        """Return True if the event matched the filter and should be used, or False otherwise."""


class RawEvent(EventFilter):
    """Matches raw core events.

    :param types: The types of event to match.
    :param func: A Callable that should accept the CoreEvent as parameter, and return
                 a boolean value indicating whether the event should be dispatched or not.
    """

    def __init__(
        self,
        types: Union[None, EventType, Iterable[EventType]] = None,
        func: Optional[Callable[[CoreEvent], bool]] = None,
    ):
        super().__init__(func=func)
        try:
            self.types = _tuple_of(types, EventType)
        except TypeError as err:
            raise TypeError(f"Invalid event type given: {types}") from err

    def __hash__(self) -> int:
        return hash((self.types, self.func))

    def __eq__(self, other) -> bool:
        if isinstance(other, RawEvent):
            return (self.types, self.func) == (other.types, other.func)
        return False

    def filter(self, event: Union[CoreEvent, NewMsgEvent]) -> bool:
        """Return True if the event matched the filter and should be used, or False otherwise."""
        assert isinstance(event, CoreEvent), "event must be an instance of CoreEvent"
        if self.types and event.kind not in self.types:
            return False
        return self._call_func(event)


class NewMessage(EventFilter):
    """Matches whenever a new message arrives. This event is only triggered by Bot clients.

    :param pattern: if set, this Pattern will be used to filter the message by its text
                    content.
    :param command: If set, only match messages with the given command (ex. /help).
                    Setting this property implies `is_info==False`.
    :param is_bot: If set to True only match messages sent by bots, if set to None
                   match messages from bots and users. If omitted or set to False
                   only messages from users will be matched.
    :param is_outgoing: If set to True only match outgoing messages, if set to None
                   match both incoming and outgoing messages. If omitted or set to
                   False only incoming messages will be matched.
    :param is_info: If set to True only match info/system messages, if set to False
                    only match messages that are not info/system messages. If omitted
                    info/system messages as well as normal messages will be matched.
    :param func: A Callable that should accept the NewMsgEvent event as parameter,
                 and return a boolean indicating whether the event should be dispatched or not.
    """

    def __init__(
        self,
        pattern: Union[
            None,
            str,
            Callable[[str], bool],
            re.Pattern,
        ] = None,
        command: Optional[str] = None,
        is_bot: Optional[bool] = False,
        is_outgoing: Optional[bool] = False,
        is_info: Optional[bool] = None,
        func: Optional[Callable[[NewMsgEvent], bool]] = None,
    ) -> None:
        super().__init__(func=func)
        self.is_bot = is_bot
        self.is_outgoing = is_outgoing
        self.is_info = is_info
        if command is not None and not isinstance(command, str):
            raise TypeError("Invalid command")
        self.command = command
        if self.is_info and self.command:
            raise AttributeError("Can not use command and is_info at the same time.")
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        if isinstance(pattern, re.Pattern):
            self.pattern: Optional[Callable] = pattern.match
        elif not pattern or callable(pattern):
            self.pattern = pattern
        else:
            raise TypeError("Invalid pattern type")

    def __hash__(self) -> int:
        return hash(
            (self.pattern, self.command, self.is_bot, self.is_outgoing, self.is_info, self.func)
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, NewMessage):
            return (
                self.pattern,
                self.command,
                self.is_bot,
                self.is_outgoing,
                self.is_info,
                self.func,
            ) == (
                other.pattern,
                other.command,
                other.is_bot,
                other.is_outgoing,
                other.is_info,
                other.func,
            )
        return False

    def filter(self, event: Union[CoreEvent, NewMsgEvent]) -> bool:
        """Return True if the event matched the filter and should be used, or False otherwise."""
        assert isinstance(event, NewMsgEvent), "event must be an instance of NewMsgEvent"
        if self.is_bot is not None and self.is_bot != event.msg.is_bot:
            return False
        if self.is_outgoing is not None and self.is_outgoing != (
            event.msg.from_id == SpecialContactId.SELF
        ):
            return False
        if self.is_info is not None and self.is_info != event.msg.is_info:
            return False
        if self.command and self.command != event.command:
            return False
        if self.pattern:
            match = self.pattern(event.msg.text)
            if not match:
                return False
        return super()._call_func(event)


_HookSet = Set[Tuple[HookCallback, Union[type, EventFilter]]]


class HookCollection:
    """
    Helper class to collect event hooks and post-hooks
    that can be later added to a Delta Chat client.
    """

    def __init__(self) -> None:
        self._hooks: _HookSet = set()
        self._post_hooks: _HookSet = set()

    def __iter__(self) -> Iterator[Tuple[HookCallback, Union[type, EventFilter]]]:
        return iter(self._hooks)

    def post_hooks_iter(self) -> Iterator[Tuple[HookCallback, Union[type, EventFilter]]]:
        """Iterator over the registered post-hooks"""
        return iter(self._post_hooks)

    def on(self, event: Union[type, EventFilter]) -> HookDecorator:
        """Register decorated function to be called for events that match the given filter."""
        return self._on(self._hooks, event)

    def after(self, event: Union[type, EventFilter]) -> HookDecorator:
        """Register decorated function to be called after an event that matches
        the given filter is processed.
        """
        return self._on(self._post_hooks, event)

    def _on(self, hooks: _HookSet, event: Union[type, EventFilter]) -> HookDecorator:
        if isinstance(event, type):
            event = event()
        assert isinstance(event, EventFilter), "Invalid event filter"

        def _decorator(func: HookCallback) -> HookCallback:
            hooks.add((func, event))
            return func

        return _decorator



"""Event loop implementations offering high level event handling/hooking."""

import logging
from typing import Callable, Dict, Iterable, Optional, Set, Tuple, Type, Union

from .events import EventFilter, HookCallback, HookCollection, RawEvent
from .rpc import Rpc
from .types import CoreEvent, Event

_HooksIndex = Dict[type, Set[Tuple[HookCallback, EventFilter]]]


class Client:
    """Delta Chat client that listen to raw core events for multiple account."""

    def __init__(
        self,
        rpc: Rpc,
        hooks: Optional[Iterable[Tuple[HookCallback, Union[type, EventFilter]]]] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """If hooks is an instance of HookCollection, also its post-hooks will be registered."""
        self.rpc = rpc
        self.logger = logger or logging.getLogger("deltachat2.Client")
        self._hooks: _HooksIndex = {}
        self._post_hooks: _HooksIndex = {}
        self.add_hooks(hooks or [])
        if isinstance(hooks, HookCollection):
            self.add_post_hooks(hooks.post_hooks_iter())

    def add_hooks(self, hooks: Iterable[Tuple[HookCallback, Union[type, EventFilter]]]) -> None:
        """Register event hooks callbacks."""
        for hook, event in hooks:
            self.add_hook(hook, event)

    def add_post_hooks(
        self, hooks: Iterable[Tuple[HookCallback, Union[type, EventFilter]]]
    ) -> None:
        """Register event post-hooks callbacks."""
        for hook, event in hooks:
            self.add_post_hook(hook, event)

    def add_hook(self, hook: HookCallback, event: Union[type, EventFilter] = RawEvent) -> None:
        """Register hook to be called when an event that matches the given filter is received."""
        event2 = event() if isinstance(event, type) else event
        self._hooks.setdefault(type(event2), set()).add((hook, event2))

    def remove_hook(self, hook: HookCallback, event: Union[type, EventFilter]) -> None:
        """Unregister hook from the given event filter."""
        event2 = event() if isinstance(event, type) else event
        self._hooks.get(type(event2), set()).remove((hook, event2))

    def add_post_hook(self, hook: HookCallback, event: Union[type, EventFilter] = RawEvent) -> None:
        """Register hook to be called after an event that matches the given filter is processed."""
        event2 = event() if isinstance(event, type) else event
        self._post_hooks.setdefault(type(event2), set()).add((hook, event2))

    def remove_post_hook(self, hook: HookCallback, event: Union[type, EventFilter]) -> None:
        """Unregister post-hook from the given event filter."""
        event2 = event() if isinstance(event, type) else event
        self._post_hooks.get(type(event2), set()).remove((hook, event2))

    def configure(self, account_id: int, email: str, password: str, **kwargs) -> None:
        """Configure the account with the given account ID."""
        self.rpc.set_config(account_id, "addr", email)
        self.rpc.set_config(account_id, "mail_pw", password)
        if kwargs:
            self.rpc.batch_set_config(account_id, kwargs)
        self.rpc.configure(account_id)

    def run_forever(self, account_id: int = 0) -> None:
        """Process events forever.

        if account_id != 0, only the account with the given ID will be started.
        """
        self.run_until(lambda _: False, account_id)

    def run_until(self, func: Callable[[Event], bool], account_id: int = 0) -> Event:
        """Process events until the given callable evaluates to True.

        The callable will receive the Event object representing the last processed event.
        The event is returned when the callable evaluates to True.
        """
        self.logger.debug("Listening to incoming events...")
        if account_id:
            if self.rpc.is_configured(account_id):
                self.rpc.start_io(account_id)
        else:
            self.rpc.start_io_for_all_accounts()

        while True:
            raw_event = self.rpc.get_next_event()
            event = Event(raw_event.context_id, CoreEvent(raw_event.event))
            self._on_event(event, RawEvent)
            if func(event):
                return event

    def _on_event(self, event: Event, filter_type: Type[EventFilter]) -> None:
        self._notify(self._hooks, event, filter_type)
        self._notify(self._post_hooks, event, filter_type)

    def _notify(self, hooks: _HooksIndex, event: Event, filter_type: Type[EventFilter]) -> None:
        for hook, evfilter in hooks.get(filter_type, []):
            if evfilter.filter(event.event):
                try:
                    hook(self, event.account_id, event.event)
                except Exception as ex:
                    self.logger.exception(ex)



"""JSON-RPC API definition."""

import dataclasses

from ._utils import _snake2camel
from .transport import RpcTransport
from .types import MsgData


class Rpc:
    """Access to the Delta Chat JSON-RPC API."""

    def __init__(self, transport: RpcTransport) -> None:
        self.transport = transport

    def __getattr__(self, attr: str):
        return lambda *args: self.transport.call(attr, *args)

    def send_msg(self, accid: int, chatid: int, data: MsgData) -> int:
        """Send a message and return the message ID of the sent message."""
        json_obj = _snake2camel(dataclasses.asdict(data))
        msgid = self.transport.call("send_msg", accid, chatid, json_obj)
        self.markseen_msgs(accid, [msgid])
        return msgid



"""Delta Chat client library using deltachat-rpc-server"""

# pylama:ignore=W0611,W0401
from .bot import Bot
from .client import Client
from .rpc import Rpc
from .transport import IOTransport, JsonRpcError, RpcTransport
from .types import *
from .util import run_bot_cli, run_client_cli



"""Minimal/simple single-account echo-bot example."""

import sys

from deltachat2 import Bot, CoreEvent, IOTransport, MsgData, NewMsgEvent, Rpc, events

hooks = events.HookCollection()


@hooks.on(events.RawEvent)
def log_event(bot: Bot, accid: int, event: CoreEvent) -> None:
    """Log all core events for debugging."""
    bot.logger.info(f"[accid={accid}] {event}")


@hooks.on(events.NewMessage)
def echo(bot: Bot, accid: int, event: NewMsgEvent) -> None:
    """Echo back any text message"""
    msg = event.msg
    bot.rpc.send_msg(accid, msg.chat_id, MsgData(text=msg.text))


def main() -> None:
    """Configure (if necessary) and run the bot."""
    with IOTransport() as trans:
        rpc = Rpc(trans)
        bot = Bot(rpc, hooks)

        accounts = rpc.get_all_account_ids()
        accid = accounts[0] if accounts else rpc.add_account()

        print("Running deltachat core", rpc.get_system_info().deltachat_core_version)

        if not rpc.is_configured(accid):
            if len(sys.argv) != 3:
                print("ERROR: Account is not configured so email and password must be provided")
                return
            bot.configure(accid, sys.argv[1], sys.argv[2])

        bot.run_forever()


if __name__ == "__main__":
    main()



"""Utilities"""

import argparse
import sys
from threading import Thread
from typing import Callable, Iterable, Optional, Tuple, Type, Union

from .bot import Bot
from .client import Client
from .events import EventFilter
from .rpc import Rpc
from .transport import IOTransport


def run_client_cli(
    hooks: Optional[Iterable[Tuple[Callable, Union[type, EventFilter]]]] = None,
    argv: Optional[list] = None,
    **kwargs,
) -> None:
    """Run a simple command line app, using the given hooks.

    Extra keyword arguments are passed to the internal Rpc object.
    """
    _run_cli(Client, hooks, argv, **kwargs)


def run_bot_cli(
    hooks: Optional[Iterable[Tuple[Callable, Union[type, EventFilter]]]] = None,
    argv: Optional[list] = None,
    **kwargs,
) -> None:
    """Run a simple bot command line using the given hooks.

    Extra keyword arguments are passed to the internal Rpc object.
    """
    _run_cli(Bot, hooks, argv, **kwargs)


def _run_cli(
    client_type: Type["Client"],
    hooks: Optional[Iterable[Tuple[Callable, Union[type, EventFilter]]]] = None,
    argv: Optional[list] = None,
    **kwargs,
) -> None:
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(prog=argv[0] if argv else None)
    parser.add_argument(
        "accounts_dir",
        help="accounts folder (default: current working directory)",
        nargs="?",
    )
    parser.add_argument("--email", action="store", help="email address")
    parser.add_argument("--password", action="store", help="password")
    args = parser.parse_args(argv[1:])

    with IOTransport(accounts_dir=args.accounts_dir, **kwargs) as trans:
        rpc = Rpc(trans)
        client = client_type(rpc, hooks)

        accounts = rpc.get_all_account_ids()
        accid = accounts[0] if accounts else rpc.add_account()

        core_version = rpc.get_system_info().deltachat_core_version
        client.logger.debug("Running deltachat core %s", core_version)

        if not rpc.is_configured(accid):
            assert args.email, "Account is not configured and email must be provided"
            assert args.password, "Account is not configured and password must be provided"
            configure_thread = Thread(
                target=client.configure, args=(accid, args.email, args.password)
            )
            configure_thread.start()

        client.run_forever()
